@@ -1,4 +1,4 @@
-/* Copyright (C) 2002-2016 Free Software Foundation, Inc.
+/* Copyright (C) 2002-2017 Free Software Foundation, Inc.
    Contributed by Andy Vaught
    F2003 I/O support contributed by Jerry DeLisle
 
@@ -27,7 +27,6 @@ see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
 
 #include "io.h"
 #include "unix.h"
-#include <stdlib.h>
 #include <limits.h>
 
 #ifdef HAVE_UNISTD_H
@@ -36,7 +35,6 @@ see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
 
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <assert.h>
 
 #include <string.h>
 #include <errno.h>
@@ -214,11 +212,11 @@ unix_stream;
 
 
 /* fix_fd()-- Given a file descriptor, make sure it is not one of the
- * standard descriptors, returning a non-standard descriptor.  If the
- * user specifies that system errors should go to standard output,
- * then closes standard output, we don't want the system errors to a
- * file that has been given file descriptor 1 or 0.  We want to send
- * the error to the invalid descriptor. */
+   standard descriptors, returning a non-standard descriptor.  If the
+   user specifies that system errors should go to standard output,
+   then closes standard output, we don't want the system errors to a
+   file that has been given file descriptor 1 or 0.  We want to send
+   the error to the invalid descriptor. */
 
 static int
 fix_fd (int fd)
@@ -262,7 +260,7 @@ fix_fd (int fd)
    corresponding C stream.  This is bugware for mixed C-Fortran codes
    where the C code doesn't flush I/O before returning.  */
 void
-flush_if_preconnected (stream * s)
+flush_if_preconnected (stream *s)
 {
   int fd;
 
@@ -289,7 +287,7 @@ than size_t as for POSIX read/write.
 *********************************************************************/
 
 static int
-raw_flush (unix_stream * s  __attribute__ ((unused)))
+raw_flush (unix_stream *s  __attribute__ ((unused)))
 {
   return 0;
 }
@@ -300,7 +298,7 @@ raw_flush (unix_stream * s  __attribute__ ((unused)))
 #define MAX_CHUNK 2147479552
 
 static ssize_t
-raw_read (unix_stream * s, void * buf, ssize_t nbyte)
+raw_read (unix_stream *s, void *buf, ssize_t nbyte)
 {
   /* For read we can't do I/O in a loop like raw_write does, because
      that will break applications that wait for interactive I/O.  We
@@ -341,7 +339,7 @@ raw_read (unix_stream * s, void * buf, ssize_t nbyte)
 }
 
 static ssize_t
-raw_write (unix_stream * s, const void * buf, ssize_t nbyte)
+raw_write (unix_stream *s, const void *buf, ssize_t nbyte)
 {
   ssize_t trans, bytes_left;
   char *buf_st;
@@ -357,7 +355,7 @@ raw_write (unix_stream * s, const void * buf, ssize_t nbyte)
     {
       ssize_t to_write = bytes_left < MAX_CHUNK ? bytes_left: MAX_CHUNK;
       trans = write (s->fd, buf_st, to_write);
-      if (trans < 0)
+      if (trans == -1)
 	{
 	  if (errno == EINTR)
 	    continue;
@@ -372,24 +370,35 @@ raw_write (unix_stream * s, const void * buf, ssize_t nbyte)
 }
 
 static gfc_offset
-raw_seek (unix_stream * s, gfc_offset offset, int whence)
+raw_seek (unix_stream *s, gfc_offset offset, int whence)
 {
-  return lseek (s->fd, offset, whence);
+  while (true)
+    {
+      gfc_offset off = lseek (s->fd, offset, whence);
+      if (off == (gfc_offset) -1 && errno == EINTR)
+	continue;
+      return off;
+    }
 }
 
 static gfc_offset
-raw_tell (unix_stream * s)
+raw_tell (unix_stream *s)
 {
-  return lseek (s->fd, 0, SEEK_CUR);
+  while (true)
+    {
+      gfc_offset off = lseek (s->fd, 0, SEEK_CUR);
+      if (off == (gfc_offset) -1 && errno == EINTR)
+	continue;
+      return off;
+    }
 }
 
 static gfc_offset
-raw_size (unix_stream * s)
+raw_size (unix_stream *s)
 {
   struct stat statbuf;
-  int ret = fstat (s->fd, &statbuf);
-  if (ret == -1)
-    return ret;
+  if (TEMP_FAILURE_RETRY (fstat (s->fd, &statbuf)) == -1)
+    return -1;
   if (S_ISREG (statbuf.st_mode))
     return statbuf.st_size;
   else
@@ -397,7 +406,7 @@ raw_size (unix_stream * s)
 }
 
 static int
-raw_truncate (unix_stream * s, gfc_offset length)
+raw_truncate (unix_stream *s, gfc_offset length)
 {
 #ifdef __MINGW32__
   HANDLE h;
@@ -431,7 +440,9 @@ raw_truncate (unix_stream * s, gfc_offset length)
   lseek (s->fd, cur, SEEK_SET);
   return -1;
 #elif defined HAVE_FTRUNCATE
-  return ftruncate (s->fd, length);
+  if (TEMP_FAILURE_RETRY (ftruncate (s->fd, length)) == -1)
+    return -1;
+  return 0;
 #elif defined HAVE_CHSIZE
   return chsize (s->fd, length);
 #else
@@ -441,7 +452,7 @@ raw_truncate (unix_stream * s, gfc_offset length)
 }
 
 static int
-raw_close (unix_stream * s)
+raw_close (unix_stream *s)
 {
   int retval;
   
@@ -450,7 +461,17 @@ raw_close (unix_stream * s)
   else if (s->fd != STDOUT_FILENO
       && s->fd != STDERR_FILENO
       && s->fd != STDIN_FILENO)
-    retval = close (s->fd);
+    {
+      retval = close (s->fd);
+      /* close() and EINTR is special, as the file descriptor is
+	 deallocated before doing anything that might cause the
+	 operation to be interrupted. Thus if we get EINTR the best we
+	 can do is ignore it and continue (otherwise if we try again
+	 the file descriptor may have been allocated again to some
+	 other file).  */
+      if (retval == -1 && errno == EINTR)
+	retval = errno = 0;
+    }
   else
     retval = 0;
   free (s);
@@ -458,7 +479,7 @@ raw_close (unix_stream * s)
 }
 
 static int
-raw_markeor (unix_stream * s __attribute__ ((unused)))
+raw_markeor (unix_stream *s __attribute__ ((unused)))
 {
   return 0;
 }
@@ -476,7 +497,7 @@ static const struct stream_vtable raw_vtable = {
 };
 
 static int
-raw_init (unix_stream * s)
+raw_init (unix_stream *s)
 {
   s->st.vptr = &raw_vtable;
 
@@ -493,7 +514,7 @@ reading to writing and vice versa.
 *********************************************************************/
 
 static int
-buf_flush (unix_stream * s)
+buf_flush (unix_stream *s)
 {
   int writelen;
 
@@ -504,7 +525,7 @@ buf_flush (unix_stream * s)
     return 0;
   
   if (s->physical_offset != s->buffer_offset
-      && lseek (s->fd, s->buffer_offset, SEEK_SET) < 0)
+      && raw_seek (s, s->buffer_offset, SEEK_SET) < 0)
     return -1;
 
   writelen = raw_write (s, s->buffer, s->ndirty);
@@ -522,7 +543,7 @@ buf_flush (unix_stream * s)
 }
 
 static ssize_t
-buf_read (unix_stream * s, void * buf, ssize_t nbyte)
+buf_read (unix_stream *s, void *buf, ssize_t nbyte)
 {
   if (s->active == 0)
     s->buffer_offset = s->logical_offset;
@@ -559,7 +580,7 @@ buf_read (unix_stream * s, void * buf, ssize_t nbyte)
       to_read = nbyte - nread;
       new_logical = s->logical_offset + nread;
       if (s->physical_offset != new_logical
-          && lseek (s->fd, new_logical, SEEK_SET) < 0)
+          && raw_seek (s, new_logical, SEEK_SET) < 0)
         return -1;
       s->buffer_offset = s->physical_offset = new_logical;
       if (to_read <= BUFFER_SIZE/2)
@@ -593,7 +614,7 @@ buf_read (unix_stream * s, void * buf, ssize_t nbyte)
 }
 
 static ssize_t
-buf_write (unix_stream * s, const void * buf, ssize_t nbyte)
+buf_write (unix_stream *s, const void *buf, ssize_t nbyte)
 {
   if (s->ndirty == 0)
     s->buffer_offset = s->logical_offset;
@@ -628,7 +649,7 @@ buf_write (unix_stream * s, const void * buf, ssize_t nbyte)
 	{
 	  if (s->physical_offset != s->logical_offset)
 	    {
-	      if (lseek (s->fd, s->logical_offset, SEEK_SET) < 0)
+	      if (raw_seek (s, s->logical_offset, SEEK_SET) < 0)
 		return -1;
 	      s->physical_offset = s->logical_offset;
 	    }
@@ -653,7 +674,7 @@ buf_write (unix_stream * s, const void * buf, ssize_t nbyte)
    when writing sequential unformatted.  */
 
 static int
-buf_markeor (unix_stream * s)
+buf_markeor (unix_stream *s)
 {
   if (s->unbuffered || s->ndirty >= BUFFER_SIZE / 2)
     return buf_flush (s);
@@ -661,7 +682,7 @@ buf_markeor (unix_stream * s)
 }
 
 static gfc_offset
-buf_seek (unix_stream * s, gfc_offset offset, int whence)
+buf_seek (unix_stream *s, gfc_offset offset, int whence)
 {
   switch (whence)
     {
@@ -686,19 +707,19 @@ buf_seek (unix_stream * s, gfc_offset offset, int whence)
 }
 
 static gfc_offset
-buf_tell (unix_stream * s)
+buf_tell (unix_stream *s)
 {
   return buf_seek (s, 0, SEEK_CUR);
 }
 
 static gfc_offset
-buf_size (unix_stream * s)
+buf_size (unix_stream *s)
 {
   return s->file_length;
 }
 
 static int
-buf_truncate (unix_stream * s, gfc_offset length)
+buf_truncate (unix_stream *s, gfc_offset length)
 {
   int r;
 
@@ -711,7 +732,7 @@ buf_truncate (unix_stream * s, gfc_offset length)
 }
 
 static int
-buf_close (unix_stream * s)
+buf_close (unix_stream *s)
 {
   if (buf_flush (s) != 0)
     return -1;
@@ -732,7 +753,7 @@ static const struct stream_vtable buf_vtable = {
 };
 
 static int
-buf_init (unix_stream * s)
+buf_init (unix_stream *s)
 {
   s->st.vptr = &buf_vtable;
 
@@ -752,9 +773,9 @@ buf_init (unix_stream * s)
 *********************************************************************/
 
 char *
-mem_alloc_r (stream * strm, int * len)
+mem_alloc_r (stream *strm, int *len)
 {
-  unix_stream * s = (unix_stream *) strm;
+  unix_stream *s = (unix_stream *) strm;
   gfc_offset n;
   gfc_offset where = s->logical_offset;
 
@@ -772,9 +793,9 @@ mem_alloc_r (stream * strm, int * len)
 
 
 char *
-mem_alloc_r4 (stream * strm, int * len)
+mem_alloc_r4 (stream *strm, int *len)
 {
-  unix_stream * s = (unix_stream *) strm;
+  unix_stream *s = (unix_stream *) strm;
   gfc_offset n;
   gfc_offset where = s->logical_offset;
 
@@ -792,9 +813,9 @@ mem_alloc_r4 (stream * strm, int * len)
 
 
 char *
-mem_alloc_w (stream * strm, int * len)
+mem_alloc_w (stream *strm, int *len)
 {
-  unix_stream * s = (unix_stream *) strm;
+  unix_stream *s = (unix_stream *)strm;
   gfc_offset m;
   gfc_offset where = s->logical_offset;
 
@@ -813,9 +834,9 @@ mem_alloc_w (stream * strm, int * len)
 
 
 gfc_char4_t *
-mem_alloc_w4 (stream * strm, int * len)
+mem_alloc_w4 (stream *strm, int *len)
 {
-  unix_stream * s = (unix_stream *) strm;
+  unix_stream *s = (unix_stream *)strm;
   gfc_offset m;
   gfc_offset where = s->logical_offset;
   gfc_char4_t *result = (gfc_char4_t *) s->buffer;
@@ -836,7 +857,7 @@ mem_alloc_w4 (stream * strm, int * len)
 /* Stream read function for character(kind=1) internal units.  */
 
 static ssize_t
-mem_read (stream * s, void * buf, ssize_t nbytes)
+mem_read (stream *s, void *buf, ssize_t nbytes)
 {
   void *p;
   int nb = nbytes;
@@ -855,7 +876,7 @@ mem_read (stream * s, void * buf, ssize_t nbytes)
 /* Stream read function for chracter(kind=4) internal units.  */
 
 static ssize_t
-mem_read4 (stream * s, void * buf, ssize_t nbytes)
+mem_read4 (stream *s, void *buf, ssize_t nbytes)
 {
   void *p;
   int nb = nbytes;
@@ -874,7 +895,7 @@ mem_read4 (stream * s, void * buf, ssize_t nbytes)
 /* Stream write function for character(kind=1) internal units.  */
 
 static ssize_t
-mem_write (stream * s, const void * buf, ssize_t nbytes)
+mem_write (stream *s, const void *buf, ssize_t nbytes)
 {
   void *p;
   int nb = nbytes;
@@ -893,7 +914,7 @@ mem_write (stream * s, const void * buf, ssize_t nbytes)
 /* Stream write function for character(kind=4) internal units.  */
 
 static ssize_t
-mem_write4 (stream * s, const void * buf, ssize_t nwords)
+mem_write4 (stream *s, const void *buf, ssize_t nwords)
 {
   gfc_char4_t *p;
   int nw = nwords;
@@ -911,9 +932,9 @@ mem_write4 (stream * s, const void * buf, ssize_t nwords)
 
 
 static gfc_offset
-mem_seek (stream * strm, gfc_offset offset, int whence)
+mem_seek (stream *strm, gfc_offset offset, int whence)
 {
-  unix_stream * s = (unix_stream *) strm;
+  unix_stream *s = (unix_stream *)strm;
   switch (whence)
     {
     case SEEK_SET:
@@ -948,14 +969,14 @@ mem_seek (stream * strm, gfc_offset offset, int whence)
 
 
 static gfc_offset
-mem_tell (stream * s)
+mem_tell (stream *s)
 {
   return ((unix_stream *)s)->logical_offset;
 }
 
 
 static int
-mem_truncate (unix_stream * s __attribute__ ((unused)), 
+mem_truncate (unix_stream *s __attribute__ ((unused)), 
 	      gfc_offset length __attribute__ ((unused)))
 {
   return 0;
@@ -963,17 +984,17 @@ mem_truncate (unix_stream * s __attribute__ ((unused)),
 
 
 static int
-mem_flush (unix_stream * s __attribute__ ((unused)))
+mem_flush (unix_stream *s __attribute__ ((unused)))
 {
   return 0;
 }
 
 
 static int
-mem_close (unix_stream * s)
+mem_close (unix_stream *s)
 {
-  free (s);
-
+  if (s)
+    free (s);
   return 0;
 }
 
@@ -1047,12 +1068,12 @@ open_internal4 (char *base, int length, gfc_offset offset)
 
   s->st.vptr = &mem4_vtable;
 
-  return (stream *) s;
+  return (stream *)s;
 }
 
 
 /* fd_to_stream()-- Given an open file descriptor, build a stream
- * around it. */
+   around it. */
 
 static stream *
 fd_to_stream (int fd, bool unformatted)
@@ -1066,7 +1087,7 @@ fd_to_stream (int fd, bool unformatted)
 
   /* Get the current length of the file. */
 
-  if (fstat (fd, &statbuf) == -1)
+  if (TEMP_FAILURE_RETRY (fstat (fd, &statbuf)) == -1)
     {
       s->st_dev = s->st_ino = -1;
       s->file_length = 0;
@@ -1162,8 +1183,8 @@ tempfile_open (const char *tempdir, char **fname)
      )
     slash = "";
 
-  // Take care that the template is longer in the mktemp() branch.
-  char * template = xmalloc (tempdirlen + 23);
+  /* Take care that the template is longer in the mktemp() branch.  */
+  char *template = xmalloc (tempdirlen + 23);
 
 #ifdef HAVE_MKSTEMP
   snprintf (template, tempdirlen + 23, "%s%sgfortrantmpXXXXXX", 
@@ -1175,9 +1196,9 @@ tempfile_open (const char *tempdir, char **fname)
 #endif
 
 #if defined(HAVE_MKOSTEMP) && defined(O_CLOEXEC)
-  fd = mkostemp (template, O_CLOEXEC);
+  TEMP_FAILURE_RETRY (fd = mkostemp (template, O_CLOEXEC));
 #else
-  fd = mkstemp (template);
+  TEMP_FAILURE_RETRY (fd = mkstemp (template));
   set_close_on_exec (fd);
 #endif
 
@@ -1219,7 +1240,7 @@ tempfile_open (const char *tempdir, char **fname)
 	continue;
       }
 
-      fd = open (template, flags, S_IRUSR | S_IWUSR);
+      TEMP_FAILURE_RETRY (fd = open (template, flags, S_IRUSR | S_IWUSR));
     }
   while (fd == -1 && errno == EEXIST);
 #ifndef O_CLOEXEC
@@ -1233,11 +1254,11 @@ tempfile_open (const char *tempdir, char **fname)
 
 
 /* tempfile()-- Generate a temporary filename for a scratch file and
- * open it.  mkstemp() opens the file for reading and writing, but the
- * library mode prevents anything that is not allowed.  The descriptor
- * is returned, which is -1 on error.  The template is pointed to by 
- * opp->file, which is copied into the unit structure
- * and freed later. */
+   open it.  mkstemp() opens the file for reading and writing, but the
+   library mode prevents anything that is not allowed.  The descriptor
+   is returned, which is -1 on error.  The template is pointed to by 
+   opp->file, which is copied into the unit structure
+   and freed later. */
 
 static int
 tempfile (st_parameter_open *opp)
@@ -1286,9 +1307,9 @@ tempfile (st_parameter_open *opp)
 
 
 /* regular_file2()-- Open a regular file.
- * Change flags->action if it is ACTION_UNSPECIFIED on entry,
- * unless an error occurs.
- * Returns the descriptor, which is less than zero on error. */
+   Change flags->action if it is ACTION_UNSPECIFIED on entry,
+   unless an error occurs.
+   Returns the descriptor, which is less than zero on error. */
 
 static int
 regular_file2 (const char *path, st_parameter_open *opp, unit_flags *flags)
@@ -1396,7 +1417,7 @@ regular_file2 (const char *path, st_parameter_open *opp, unit_flags *flags)
 #endif
 
   mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
-  fd = open (path, rwflag | crflag, mode);
+  TEMP_FAILURE_RETRY (fd = open (path, rwflag | crflag, mode));
   if (flags->action != ACTION_UNSPECIFIED)
     return fd;
 
@@ -1414,7 +1435,7 @@ regular_file2 (const char *path, st_parameter_open *opp, unit_flags *flags)
     crflag2 = crflag & ~(O_CREAT);
   else
     crflag2 = crflag;
-  fd = open (path, rwflag | crflag2, mode);
+  TEMP_FAILURE_RETRY (fd = open (path, rwflag | crflag2, mode));
   if (fd >=0)
     {
       flags->action = ACTION_READ;
@@ -1426,7 +1447,7 @@ regular_file2 (const char *path, st_parameter_open *opp, unit_flags *flags)
 
   /* retry for write-only access */
   rwflag = O_WRONLY;
-  fd = open (path, rwflag | crflag, mode);
+  TEMP_FAILURE_RETRY (fd = open (path, rwflag | crflag, mode));
   if (fd >=0)
     {
       flags->action = ACTION_WRITE;
@@ -1436,6 +1457,56 @@ regular_file2 (const char *path, st_parameter_open *opp, unit_flags *flags)
 }
 
 
+/* Lock the file, if necessary, based on SHARE flags.  */
+
+#if defined(HAVE_FCNTL) && defined(F_SETLK) && defined(F_UNLCK)
+static int
+open_share (st_parameter_open *opp, int fd, unit_flags *flags)
+{
+  int r = 0;
+  struct flock f;
+  if (fd == STDOUT_FILENO || fd == STDERR_FILENO || fd == STDIN_FILENO)
+    return 0;
+
+  f.l_start = 0;
+  f.l_len = 0;
+  f.l_whence = SEEK_SET;
+
+  switch (flags->share)
+  {
+    case SHARE_DENYNONE:
+      f.l_type = F_RDLCK;
+      r = fcntl (fd, F_SETLK, &f);
+      break;
+    case SHARE_DENYRW:
+      /* Must be writable to hold write lock.  */
+      if (flags->action == ACTION_READ)
+	{
+	  generate_error (&opp->common, LIBERROR_BAD_ACTION,
+	      "Cannot set write lock on file opened for READ");
+	  return -1;
+	}
+      f.l_type = F_WRLCK;
+      r = fcntl (fd, F_SETLK, &f);
+      break;
+    case SHARE_UNSPECIFIED:
+    default:
+      break;
+  }
+
+  return r;
+}
+#else
+static int
+open_share (st_parameter_open *opp __attribute__ ((unused)),
+    int fd __attribute__ ((unused)),
+    unit_flags *flags __attribute__ ((unused)))
+{
+  return 0;
+}
+#endif /* defined(HAVE_FCNTL) ... */
+
+
 /* Wrapper around regular_file2, to make sure we free the path after
    we're done.  */
 
@@ -1449,8 +1520,8 @@ regular_file (st_parameter_open *opp, unit_flags *flags)
 }
 
 /* open_external()-- Open an external file, unix specific version.
- * Change flags->action if it is ACTION_UNSPECIFIED on entry.
- * Returns NULL on operating system error. */
+   Change flags->action if it is ACTION_UNSPECIFIED on entry.
+   Returns NULL on operating system error. */
 
 stream *
 open_external (st_parameter_open *opp, unit_flags *flags)
@@ -1461,7 +1532,7 @@ open_external (st_parameter_open *opp, unit_flags *flags)
     {
       fd = tempfile (opp);
       if (flags->action == ACTION_UNSPECIFIED)
-	flags->action = ACTION_READWRITE;
+	flags->action = flags->readonly ? ACTION_READ : ACTION_READWRITE;
 
 #if HAVE_UNLINK_OPEN_FILE
       /* We can unlink scratch files now and it will go away when closed. */
@@ -1472,7 +1543,7 @@ open_external (st_parameter_open *opp, unit_flags *flags)
   else
     {
       /* regular_file resets flags->action if it is ACTION_UNSPECIFIED and
-       * if it succeeds */
+         if it succeeds */
       fd = regular_file (opp, flags);
 #ifndef O_CLOEXEC
       set_close_on_exec (fd);
@@ -1483,12 +1554,15 @@ open_external (st_parameter_open *opp, unit_flags *flags)
     return NULL;
   fd = fix_fd (fd);
 
+  if (open_share (opp, fd, flags) < 0)
+    return NULL;
+
   return fd_to_stream (fd, flags->form == FORM_UNFORMATTED);
 }
 
 
 /* input_stream()-- Return a stream pointer to the default input stream.
- * Called on initialization. */
+   Called on initialization. */
 
 stream *
 input_stream (void)
@@ -1498,12 +1572,12 @@ input_stream (void)
 
 
 /* output_stream()-- Return a stream pointer to the default output stream.
- * Called on initialization. */
+   Called on initialization. */
 
 stream *
 output_stream (void)
 {
-  stream * s;
+  stream *s;
 
 #if defined(HAVE_CRLF) && defined(HAVE_SETMODE)
   setmode (STDOUT_FILENO, O_BINARY);
@@ -1515,12 +1589,12 @@ output_stream (void)
 
 
 /* error_stream()-- Return a stream pointer to the default error stream.
- * Called on initialization. */
+   Called on initialization. */
 
 stream *
 error_stream (void)
 {
-  stream * s;
+  stream *s;
 
 #if defined(HAVE_CRLF) && defined(HAVE_SETMODE)
   setmode (STDERR_FILENO, O_BINARY);
@@ -1532,8 +1606,8 @@ error_stream (void)
 
 
 /* compare_file_filename()-- Given an open stream and a fortran string
- * that is a filename, figure out if the file is the same as the
- * filename. */
+   that is a filename, figure out if the file is the same as the
+   filename. */
 
 int
 compare_file_filename (gfc_unit *u, const char *name, int len)
@@ -1551,9 +1625,9 @@ compare_file_filename (gfc_unit *u, const char *name, int len)
   char *path = fc_strdup (name, len);
 
   /* If the filename doesn't exist, then there is no match with the
-   * existing file. */
+     existing file. */
 
-  if (stat (path, &st) < 0)
+  if (TEMP_FAILURE_RETRY (stat (path, &st)) < 0)
     {
       ret = 0;
       goto done;
@@ -1642,7 +1716,7 @@ find_file0 (gfc_unit *u, FIND_FILE0_DECL)
 
 
 /* find_file()-- Take the current filename and see if there is a unit
- * that has the file already open.  Returns a pointer to the unit if so. */
+   that has the file already open.  Returns a pointer to the unit if so. */
 
 gfc_unit *
 find_file (const char *file, gfc_charlen_type file_len)
@@ -1655,7 +1729,7 @@ find_file (const char *file, gfc_charlen_type file_len)
 
   char *path = fc_strdup (file, file_len);
 
-  if (stat (path, &st[0]) < 0)
+  if (TEMP_FAILURE_RETRY (stat (path, &st[0])) < 0)
     {
       u = NULL;
       goto done;
@@ -1763,8 +1837,42 @@ flush_all_units (void)
 }
 
 
+/* Unlock the unit if necessary, based on SHARE flags.  */
+
+int
+close_share (gfc_unit *u __attribute__ ((unused)))
+{
+  int r = 0;
+#if defined(HAVE_FCNTL) && defined(F_SETLK) && defined(F_UNLCK)
+  unix_stream *s = (unix_stream *) u->s;
+  int fd = s->fd;
+  struct flock f;
+
+  switch (u->flags.share)
+  {
+    case SHARE_DENYRW:
+    case SHARE_DENYNONE:
+      if (fd != STDOUT_FILENO && fd != STDERR_FILENO && fd != STDIN_FILENO)
+	{
+	  f.l_start = 0;
+	  f.l_len = 0;
+	  f.l_whence = SEEK_SET;
+	  f.l_type = F_UNLCK;
+	  r = fcntl (fd, F_SETLK, &f);
+	}
+      break;
+    case SHARE_UNSPECIFIED:
+    default:
+      break;
+  }
+
+#endif
+  return r;
+}
+
+
 /* file_exists()-- Returns nonzero if the current filename exists on
- * the system */
+   the system */
 
 int
 file_exists (const char *file, gfc_charlen_type file_len)
@@ -1783,7 +1891,8 @@ file_size (const char *file, gfc_charlen_type file_len)
 {
   char *path = fc_strdup (file, file_len);
   struct stat statbuf;
-  int err = stat (path, &statbuf);
+  int err;
+  TEMP_FAILURE_RETRY (err = stat (path, &statbuf));
   free (path);
   if (err == -1)
     return -1;
@@ -1793,8 +1902,8 @@ file_size (const char *file, gfc_charlen_type file_len)
 static const char yes[] = "YES", no[] = "NO", unknown[] = "UNKNOWN";
 
 /* inquire_sequential()-- Given a fortran string, determine if the
- * file is suitable for sequential access.  Returns a C-style
- * string. */
+   file is suitable for sequential access.  Returns a C-style
+   string. */
 
 const char *
 inquire_sequential (const char *string, int len)
@@ -1805,7 +1914,8 @@ inquire_sequential (const char *string, int len)
     return unknown;
 
   char *path = fc_strdup (string, len);
-  int err = stat (path, &statbuf);
+  int err;
+  TEMP_FAILURE_RETRY (err = stat (path, &statbuf));
   free (path);
   if (err == -1)
     return unknown;
@@ -1822,7 +1932,7 @@ inquire_sequential (const char *string, int len)
 
 
 /* inquire_direct()-- Given a fortran string, determine if the file is
- * suitable for direct access.  Returns a C-style string. */
+   suitable for direct access.  Returns a C-style string. */
 
 const char *
 inquire_direct (const char *string, int len)
@@ -1833,7 +1943,8 @@ inquire_direct (const char *string, int len)
     return unknown;
 
   char *path = fc_strdup (string, len);
-  int err = stat (path, &statbuf);
+  int err;
+  TEMP_FAILURE_RETRY (err = stat (path, &statbuf));
   free (path);
   if (err == -1)
     return unknown;
@@ -1850,7 +1961,7 @@ inquire_direct (const char *string, int len)
 
 
 /* inquire_formatted()-- Given a fortran string, determine if the file
- * is suitable for formatted form.  Returns a C-style string. */
+   is suitable for formatted form.  Returns a C-style string. */
 
 const char *
 inquire_formatted (const char *string, int len)
@@ -1861,7 +1972,8 @@ inquire_formatted (const char *string, int len)
     return unknown;
 
   char *path = fc_strdup (string, len);
-  int err = stat (path, &statbuf);
+  int err;
+  TEMP_FAILURE_RETRY (err = stat (path, &statbuf));
   free (path);
   if (err == -1)
     return unknown;
@@ -1879,7 +1991,7 @@ inquire_formatted (const char *string, int len)
 
 
 /* inquire_unformatted()-- Given a fortran string, determine if the file
- * is suitable for unformatted form.  Returns a C-style string. */
+   is suitable for unformatted form.  Returns a C-style string. */
 
 const char *
 inquire_unformatted (const char *string, int len)
@@ -1889,7 +2001,7 @@ inquire_unformatted (const char *string, int len)
 
 
 /* inquire_access()-- Given a fortran string, determine if the file is
- * suitable for access. */
+   suitable for access. */
 
 static const char *
 inquire_access (const char *string, int len, int mode)
@@ -1907,7 +2019,7 @@ inquire_access (const char *string, int len, int mode)
 
 
 /* inquire_read()-- Given a fortran string, determine if the file is
- * suitable for READ access. */
+   suitable for READ access. */
 
 const char *
 inquire_read (const char *string, int len)
@@ -1917,7 +2029,7 @@ inquire_read (const char *string, int len)
 
 
 /* inquire_write()-- Given a fortran string, determine if the file is
- * suitable for READ access. */
+   suitable for READ access. */
 
 const char *
 inquire_write (const char *string, int len)
@@ -1927,7 +2039,7 @@ inquire_write (const char *string, int len)
 
 
 /* inquire_readwrite()-- Given a fortran string, determine if the file is
- * suitable for read and write access. */
+   suitable for read and write access. */
 
 const char *
 inquire_readwrite (const char *string, int len)
@@ -1944,15 +2056,15 @@ stream_isatty (stream *s)
 
 int
 stream_ttyname (stream *s  __attribute__ ((unused)),
-		char * buf  __attribute__ ((unused)),
+		char *buf  __attribute__ ((unused)),
 		size_t buflen  __attribute__ ((unused)))
 {
 #ifdef HAVE_TTYNAME_R
-  return ttyname_r (((unix_stream *) s)->fd, buf, buflen);
+  return ttyname_r (((unix_stream *)s)->fd, buf, buflen);
 #elif defined HAVE_TTYNAME
   char *p;
   size_t plen;
-  p = ttyname (((unix_stream *) s)->fd);
+  p = ttyname (((unix_stream *)s)->fd);
   if (!p)
     return errno;
   plen = strlen (p);
